# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#

import numpy

from tvb.adapters.simulator.equation_forms import get_ui_name_to_monitor_equation_dict, HRFKernelEquation
from tvb.core.entities.file.simulator.view_model import *
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.entities.load import load_entity_by_gid
from tvb.core.neotraits.forms import Form, ArrayField, MultiSelectField, FloatField, StrField
from tvb.core.neotraits.forms import SelectField, TraitDataTypeSelectField
from tvb.datatypes.projections import ProjectionsType
from tvb.datatypes.sensors import SensorTypes


def get_monitor_to_form_dict():
    monitor_class_to_form = {
        RawViewModel: MonitorForm,
        SubSampleViewModel: MonitorForm,
        SpatialAverageViewModel: SpatialAverageMonitorForm,
        GlobalAverageViewModel: MonitorForm,
        TemporalAverageViewModel: MonitorForm,
        EEGViewModel: EEGMonitorForm,
        MEGViewModel: MEGMonitorForm,
        iEEGViewModel: iEEGMonitorForm,
        BoldViewModel: BoldMonitorForm,
        BoldRegionROIViewModel: BoldMonitorForm
    }

    return monitor_class_to_form


def get_ui_name_to_monitor_dict(surface):
    ui_name_to_monitor = {
        'Raw recording': RawViewModel,
        'Temporally sub-sample': SubSampleViewModel,
        'Spatial average with temporal sub-sample': SpatialAverageViewModel,
        'Global average': GlobalAverageViewModel,
        'Temporal average': TemporalAverageViewModel,
        'EEG': EEGViewModel,
        'MEG': MEGViewModel,
        'Intracerebral / Stereo EEG': iEEGViewModel,
        'BOLD': BoldViewModel
    }

    if surface:
        ui_name_to_monitor['BOLD Region ROI'] = BoldRegionROIViewModel

    return ui_name_to_monitor


def get_monitor_to_ui_name_dict(surface):
    monitor_to_ui_name = dict((v, k) for k, v in get_ui_name_to_monitor_dict(surface).items())
    return monitor_to_ui_name


def get_form_for_monitor(monitor_class):
    return get_monitor_to_form_dict().get(monitor_class)


class MonitorForm(Form):

    def __init__(self, session_stored_simulator=None):
        super(MonitorForm, self).__init__()
        self.session_stored_simulator = session_stored_simulator
        self.period = FloatField(Monitor.period)
        self.variables_of_interest_indexes = {}

        if session_stored_simulator is not None:
            self.variables_of_interest_indexes = session_stored_simulator.determine_indexes_for_chosen_vars_of_interest()

        self.variables_of_interest = MultiSelectField(List(of=str, label='Model Variables to watch',
                                                           choices=tuple(self.variables_of_interest_indexes.keys())),
                                                      name='variables_of_interest')

    def fill_from_trait(self, trait):
        super(MonitorForm, self).fill_from_trait(trait)
        if trait.variables_of_interest is not None:
            self.variables_of_interest.data = [list(self.variables_of_interest_indexes.keys())[idx]
                                               for idx in trait.variables_of_interest]
        else:
            # by default we select all variables of interest for the monitor forms
            self.variables_of_interest.data = list(self.variables_of_interest_indexes.keys())

    def fill_trait(self, datatype):
        super(MonitorForm, self).fill_trait(datatype)
        datatype.variables_of_interest = numpy.array(list(self.variables_of_interest_indexes.values()))

    def fill_from_post(self, form_data):
        super(MonitorForm, self).fill_from_post(form_data)
        all_variables = self.session_stored_simulator.model.variables_of_interest
        chosen_variables = form_data['variables_of_interest']
        self.variables_of_interest_indexes = self.session_stored_simulator.\
            get_variables_of_interest_indexes(all_variables, chosen_variables)


class SpatialAverageMonitorForm(MonitorForm):

    def __init__(self, session_stored_simulator=None):
        super(SpatialAverageMonitorForm, self).__init__(session_stored_simulator)
        self.spatial_mask = ArrayField(SpatialAverage.spatial_mask)
        self.default_mask = SelectField(SpatialAverage.default_mask)

    def fill_from_trait(self, trait):
        super(SpatialAverageMonitorForm, self).fill_from_trait(trait)
        connectivity_index = load_entity_by_gid(self.session_stored_simulator.connectivity)

        if self.session_stored_simulator.is_surface_simulation is False:
            self.default_mask.choices.pop(SpatialAverage.REGION_MAPPING)

            if connectivity_index.has_cortical_mask is False:
                self.default_mask.choices.pop(SpatialAverage.CORTICAL)

            if connectivity_index.has_hemispheres_mask is False:
                self.default_mask.choices.pop(SpatialAverage.HEMISPHERES)

        else:
            self.default_mask.data = SpatialAverage.REGION_MAPPING
            self.default_mask.disabled = True


class ProjectionMonitorForm(MonitorForm):

    def __init__(self, session_stored_simulator=None):
        super(ProjectionMonitorForm, self).__init__(session_stored_simulator)

        rm_filter = None
        if session_stored_simulator and session_stored_simulator.is_surface_simulation:
            rm_filter = FilterChain(fields=[FilterChain.datatype + '.gid'], operations=['=='],
                                    values=[session_stored_simulator.surface.region_mapping_data.hex])

        self.region_mapping = TraitDataTypeSelectField(ProjectionViewModel.region_mapping, name='region_mapping',
                                                       conditions=rm_filter)


class EEGMonitorForm(ProjectionMonitorForm):

    def __init__(self, session_stored_simulator=None):
        super(EEGMonitorForm, self).__init__(session_stored_simulator)

        sensor_filter = FilterChain(fields=[FilterChain.datatype + '.sensors_type'], operations=["=="],
                                    values=[SensorTypes.TYPE_EEG.value])

        projection_filter = FilterChain(fields=[FilterChain.datatype + '.projection_type'], operations=["=="],
                                        values=[ProjectionsType.EEG.value])

        self.projection = TraitDataTypeSelectField(EEGViewModel.projection, name='projection',
                                                   conditions=projection_filter)
        self.reference = StrField(EEG.reference)
        self.sensors = TraitDataTypeSelectField(EEGViewModel.sensors, name='sensors', conditions=sensor_filter)
        self.sigma = FloatField(EEG.sigma)


class MEGMonitorForm(ProjectionMonitorForm):

    def __init__(self, session_stored_simulator=None):
        super(MEGMonitorForm, self).__init__(session_stored_simulator)

        sensor_filter = FilterChain(fields=[FilterChain.datatype + '.sensors_type'], operations=["=="],
                                    values=[SensorTypes.TYPE_MEG.value])

        projection_filter = FilterChain(fields=[FilterChain.datatype + '.projection_type'], operations=["=="],
                                        values=[ProjectionsType.MEG.value])

        self.projection = TraitDataTypeSelectField(MEGViewModel.projection, name='projection',
                                                   conditions=projection_filter)
        self.sensors = TraitDataTypeSelectField(MEGViewModel.sensors, name='sensors', conditions=sensor_filter)


class iEEGMonitorForm(ProjectionMonitorForm):

    def __init__(self, session_stored_simulator=None):
        super(iEEGMonitorForm, self).__init__(session_stored_simulator)

        sensor_filter = FilterChain(fields=[FilterChain.datatype + '.sensors_type'], operations=["=="],
                                    values=[SensorTypes.TYPE_INTERNAL.value])

        projection_filter = FilterChain(fields=[FilterChain.datatype + '.projection_type'], operations=["=="],
                                        values=[ProjectionsType.SEEG.value])

        self.projection = TraitDataTypeSelectField(iEEGViewModel.projection, name='projection',
                                                   conditions=projection_filter)
        self.sigma = FloatField(iEEG.sigma)
        self.sensors = TraitDataTypeSelectField(iEEGViewModel.sensors, name='sensors', conditions=sensor_filter)


class BoldMonitorForm(MonitorForm):

    def __init__(self, session_stored_simulator=None):
        super(BoldMonitorForm, self).__init__(session_stored_simulator)
        self.hrf_kernel_choices = get_ui_name_to_monitor_equation_dict()
        default_hrf_kernel = list(self.hrf_kernel_choices.values())[0]

        self.period = FloatField(Bold.period)
        self.hrf_kernel = SelectField(Attr(HRFKernelEquation, label='Equation', default=default_hrf_kernel),
                                      name='hrf_kernel', choices=self.hrf_kernel_choices)

    def fill_trait(self, datatype):
        super(BoldMonitorForm, self).fill_trait(datatype)
        datatype.period = self.period.data
        if type(datatype.hrf_kernel) != self.hrf_kernel.data:
            datatype.hrf_kernel = self.hrf_kernel.data()

    def fill_from_trait(self, trait):
        super(BoldMonitorForm, self).fill_from_trait(trait)
        self.hrf_kernel.data = trait.hrf_kernel.__class__



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
A matrix visualizer for the Independent Component Analysis.
It displays the mixing matrix of size n_features x n_components

.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
.. moduleauthor:: Paula Sanz Leon <Paula@tvb.invalid>
"""

from tvb.adapters.datatypes.db.mode_decompositions import IndependentComponentsIndex
from tvb.adapters.visualizers.matrix_viewer import ABCMappedArraySVGVisualizer
from tvb.basic.neotraits.api import Attr
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.core.adapters.arguments_serialisation import slice_str
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import TraitDataTypeSelectField, IntField
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr
from tvb.datatypes.mode_decompositions import IndependentComponents


class ICAModel(ViewModel):
    datatype = DataTypeGidAttr(
        linked_datatype=IndependentComponents,
        label='Independent component analysis:'
    )

    i_svar = Attr(
        field_type=int,
        default=0,
        label='Index of state variable (defaults to first state variable)'
    )

    i_mode = Attr(
        field_type=int,
        default=0,
        label='Index of mode (defaults to first mode)'
    )


class ICAForm(ABCAdapterForm):

    def __init__(self):
        super(ICAForm, self).__init__()
        self.datatype = TraitDataTypeSelectField(ICAModel.datatype, name='datatype', conditions=self.get_filters())
        self.i_svar = IntField(ICAModel.i_svar, name='i_svar')
        self.i_mode = IntField(ICAModel.i_mode, name='i_mode')

    @staticmethod
    def get_view_model():
        return ICAModel

    @staticmethod
    def get_required_datatype():
        return IndependentComponentsIndex

    @staticmethod
    def get_filters():
        return None

    @staticmethod
    def get_input_name():
        return 'datatype'


class ICA(ABCMappedArraySVGVisualizer):
    _ui_name = "Independent Components Analysis Visualizer"
    _ui_subsection = "ica"

    def get_form_class(self):
        return ICAForm

    def launch(self, view_model):
        # type: (ICAModel) -> dict
        """Construct data for visualization and launch it."""
        ica_gid = view_model.datatype
        ica_index = self.load_entity_by_gid(ica_gid)

        slice_given = slice_str((slice(None), slice(None), slice(view_model.i_svar), slice(view_model.i_mode)))
        if view_model.i_svar < 0 or view_model.i_svar >= ica_index.parsed_shape[2]:
            view_model.i_svar = 0
        if view_model.i_mode < 0 or view_model.i_mode >= ica_index.parsed_shape[3]:
            view_model.i_mode = 0
        slice_used = slice_str((slice(None), slice(None), slice(view_model.i_svar), slice(view_model.i_mode)))

        with h5.h5_file_for_index(ica_index) as h5_file:
            unmixing_matrix = h5_file.unmixing_matrix[..., view_model.i_svar, view_model.i_mode]
            prewhitening_matrix = h5_file.prewhitening_matrix[..., view_model.i_svar, view_model.i_mode]
        Cinv = unmixing_matrix.dot(prewhitening_matrix)

        title = 'ICA region contribution -- (Ellipsis, %d, 0)' % (view_model.i_svar)
        labels = self.extract_source_labels(ica_index)
        pars = self.compute_params(ica_index, Cinv, title, [labels, labels],
                                   slice_given, slice_used, slice_given != slice_used)
        return self.build_display_result("matrix/svg_view", pars)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""
import numpy
import six
from tvb.basic.logger.builder import get_logger

# To plot nullclines we need a function that computes contours of a scalar field.
# We use the internal matplotlib one.
# Now that newer matplotlib versions have changed this internal API it would be a
# good idea to use a proper library for this.

try:
    from matplotlib import _cntr
    # older matplotlib

    def nullcline(x, y, z):
        c = _cntr.Cntr(x, y, z)
        # trace a contour
        res = c.trace(0.0)
        if not res:
            return numpy.array([])
        # result is a list of arrays of vertices and path codes
        # (see docs for matplotlib.path.Path)
        nseg = len(res) // 2
        segments, codes = res[:nseg], res[nseg:]
        return segments

except ImportError:
    from matplotlib import _contour
    # newer matplotlib >= 2.2

    def nullcline(x, y, z):
        c = _contour.QuadContourGenerator(x, y, z, None, True, 0)
        segments = c.create_contour(0.0)
        return segments

# how much courser is the grid used to show the vectors
GRID_SUBSAMPLE = 2
# Set the resolution of the phase-plane and sample trajectories.
NUMBEROFGRIDPOINTS = GRID_SUBSAMPLE * 42


class _PhaseSpace(object):
    """
    Dimensionality independent code
    """

    def __init__(self, model, integrator):
        self.log = get_logger(self.__class__.__module__)
        self.model = model
        self.integrator = integrator


    def _compute_trajectories(self, states, n_steps):
        """
        A vectorized method of computing a number of trajectories in parallel.
        Returns a collection of nvar-dimensional trajectories.
        """
        scheme = self.integrator.scheme
        trajs = numpy.zeros((n_steps + 1, self.model.nvar, len(states), self.model.number_of_modes))
        # reshape to what dfun expects: from n, sv to sv, n, mode
        states = numpy.tile(states.T[:, :, numpy.newaxis], self.model.number_of_modes)
        trajs[0, :] = states
        no_coupling = numpy.zeros((self.model.nvar, states.shape[1], self.model.number_of_modes))
        # grow trajectories step by step
        for step in range(n_steps):
            states = scheme(states, self.model.dfun, no_coupling, 0.0, 0.0)
            trajs[step + 1, :] = states

        if numpy.isnan(trajs).any():
            self.log.warn("NaN in trajectories")
        return trajs


    def get_axes_ranges(self, sv):
        lo, hi = self.model.state_variable_range[sv]
        default_range = hi - lo
        min_val = lo - 4.0 * default_range
        max_val = hi + 4.0 * default_range
        return min_val, max_val, lo, hi


class PhasePlane(_PhaseSpace):
    """
    Responsible with computing phase space slices and trajectories.
    A collection of math-y utilities it is view independent (holds no state related to views).
    """

    @staticmethod
    def _create_mesh_jitter():
        shape = NUMBEROFGRIDPOINTS, NUMBEROFGRIDPOINTS
        d = 1.0 / (4 * NUMBEROFGRIDPOINTS)
        return numpy.random.normal(0, d, shape), numpy.random.normal(0, d, shape)


    def _get_mesh_grid(self, svx_ind, svy_ind, noise=None):
        """
        Generate the phase-plane gridding based on the given
        state-variable indices and their range values.
        """
        svr = self.model.state_variable_range
        xlo, xhi = svr[self.model.state_variables[svx_ind]]
        ylo, yhi = svr[self.model.state_variables[svy_ind]]

        xg = numpy.mgrid[xlo:xhi:(NUMBEROFGRIDPOINTS * 1j)]
        yg = numpy.mgrid[ylo:yhi:(NUMBEROFGRIDPOINTS * 1j)]
        xgr, ygr = numpy.meshgrid(xg, yg)
        if noise:
            # add scaled noise
            xgr += noise[0] * (xhi - xlo)
            ygr += noise[1] * (yhi - ylo)
        return xgr, ygr


    def _calc_phase_plane(self, state, svx_ind, svy_ind, xg, yg):
        """
        Computes a 2d axis aligned rectangle of the vector field returning a u, v vector field.
        The slice passes through the `state` point and varies along the axes given by svx_ind and svy_ind.
        The last 2 parameters specify the mesh in the varying directions. To be computed by _get_mesh_grid
        Vectorized function, it evaluate all grid points at once as if they were connectivity nodes.
        """
        state_variables = numpy.tile(state, (NUMBEROFGRIDPOINTS ** 2, 1))

        for mode_idx in range(self.model.number_of_modes):
            state_variables[svx_ind, :, mode_idx] = xg.flat
            state_variables[svy_ind, :, mode_idx] = yg.flat

        no_coupling = numpy.zeros((self.model.nvar, state_variables.shape[1], self.model.number_of_modes))
        d_grid = self.model.dfun(state_variables, no_coupling)

        flat_uv_grid = d_grid[[svx_ind, svy_ind], :, :]  # subset of the state variables to be displayed
        u, v = flat_uv_grid.reshape((2, NUMBEROFGRIDPOINTS, NUMBEROFGRIDPOINTS, self.model.number_of_modes))
        if numpy.isnan(u).any() or numpy.isnan(v).any():
            self.log.error("NaN")
        return u, v



class PhasePlaneD3(PhasePlane):
    """
    Provides data for a d3 client
    """

    def __init__(self, model, integrator):
        PhasePlane.__init__(self, model, integrator)
        self.mode = 0
        self.svx_ind = 0  # x-axis: 1st state variable
        if self.model.nvar > 1:
            self.svy_ind = 1  # y-axis: 2nd state variable
        else:
            self.svy_ind = 0
        # Set up a vector containing the default state-variable values
        svr = self.model.state_variable_range
        sv_mean = numpy.array([svr[key].mean() for key in self.model.state_variables])
        sv_mean = sv_mean.reshape((self.model.nvar, 1, 1))
        self.default_sv = sv_mean.repeat(self.model.number_of_modes, axis=2)
        self.update_integrator_clamping()
        self._jitter = None  # self._create_mesh_jitter()

    def update_integrator_clamping(self):
        clamped_sv_indices = [i for i in range(self.model.nvar) if i not in [self.svx_ind, self.svy_ind]]
        if clamped_sv_indices:
            self.integrator.clamped_state_variable_indices = numpy.array(clamped_sv_indices)
            self.integrator.clamped_state_variable_values = self.default_sv[
                self.integrator.clamped_state_variable_indices]
        else:
            self.integrator.clamped_state_variable_indices = None
            self.integrator.clamped_state_variable_values = None

    def update_axis(self, mode, svx, svy, x_range, y_range, state_vars):
        self.mode = int(mode)
        self.svx_ind = self.model.state_variables.index(svx)
        self.svy_ind = self.model.state_variables.index(svy)
        svr = self.model.state_variable_range
        svr[svx][:] = x_range  # todo is it ok to update the model?
        svr[svy][:] = y_range

        for name, val in six.iteritems(state_vars):
            k = self.model.state_variables.index(name)
            self.default_sv[k] = val
        self.update_integrator_clamping()


    def compute_phase_plane(self):
        """
        :return: A json representation of the phase plane.
        """
        x, y = self._get_mesh_grid(self.svx_ind, self.svy_ind, noise=self._jitter)

        u, v = self._calc_phase_plane(self.default_sv, self.svx_ind, self.svy_ind, x, y)
        u = u[..., self.mode]  # project on active mode
        v = v[..., self.mode]
        xnull = [{'path': segment.tolist(), 'nullcline_index': 0} for segment in nullcline(x, y, u)]
        ynull = [{'path': segment.tolist(), 'nullcline_index': 1} for segment in nullcline(x, y, v)]

        # a courser mesh for the arrows
        xsmall = x[::GRID_SUBSAMPLE, ::GRID_SUBSAMPLE]
        ysmall = y[::GRID_SUBSAMPLE, ::GRID_SUBSAMPLE]
        usmall = u[::GRID_SUBSAMPLE, ::GRID_SUBSAMPLE]
        vsmall = v[::GRID_SUBSAMPLE, ::GRID_SUBSAMPLE]

        d = numpy.dstack((xsmall, ysmall, usmall, vsmall))
        d = d.reshape(((NUMBEROFGRIDPOINTS // GRID_SUBSAMPLE) ** 2, 4)).tolist()

        return {'plane': d, 'nullclines': xnull + ynull}


    def _state_dict_to_array(self, state):
        arr = numpy.zeros(len(self.model.state_variables))
        for svn, v in six.iteritems(state):
            svn_idx = self.model.state_variables.index(svn)
            arr[svn_idx] = v
        return arr


    def trajectories(self, starting_points, n_steps=512):
        """
        :param starting_points: A list of starting points represented as dicts of state_var_name to value
        :return: a tuple of trajectories and signals
        """
        starting_points = numpy.array([self._state_dict_to_array(s) for s in starting_points])
        traj = self._compute_trajectories(starting_points, n_steps)  # point_on_traj_idx, sv_idx, traj_idx, mode
        # reshape it and project it on the plane defined  by the current axis state vars
        traj = traj.transpose(2, 0, 1, 3)  # traj_idx, point, sv_idx, mode
        trajectory = traj[:, :, [self.svx_ind, self.svy_ind], self.mode]  # traj_idx, points, x, y

        # signals for last trajectory
        signal_x = numpy.arange(n_steps + 1) * self.integrator.dt
        signals = [list(zip(signal_x, traj[-1, :, i, self.mode].tolist())) for i in [self.svx_ind, self.svy_ind]]

        return trajectory.tolist(), signals


class PhaseLineD3(_PhaseSpace):
    def __init__(self, model, integrator):
        _PhaseSpace.__init__(self, model, integrator)
        self.mode = 0


    def _grid(self):
        svr = self.model.state_variable_range
        xlo, xhi = svr[self.model.state_variables[0]]
        return numpy.linspace(xlo, xhi, NUMBEROFGRIDPOINTS)


    def compute_phase_plane(self):
        xg = self._grid()
        # dfun modifies state in place so we need to copy xg
        state = xg.reshape((1, NUMBEROFGRIDPOINTS, 1)).copy()  # will broadcast to modes
        no_coupling = numpy.zeros((self.model.nvar, state.shape[1], self.model.number_of_modes))
        u = self.model.dfun(state, no_coupling)
        u = u[0, :, self.mode]

        d = numpy.vstack((xg, u)).T

        if numpy.isnan(d).any():
            self.log.error("NaN")

        # find zeroes. This method is not exact
        zero_crossings = numpy.where(numpy.diff(numpy.sign(u)))[0]
        zero_crossings = xg[zero_crossings]
        return {'signal': d.tolist(), 'zeroes': zero_crossings.tolist()}


    def update_axis(self, mode, svx, x_range):
        self.mode = int(mode)
        svr = self.model.state_variable_range
        svr[svx][:] = x_range


def phase_space_d3(model, integrator):
    """
    :return: A phase plane or a phase line depending on the dimensionality of the model
    """
    if model.nvar == 1:
        return PhaseLineD3(model, integrator)
    else:
        return PhasePlaneD3(model, integrator)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from sqlalchemy import Column, Integer, ForeignKey, String, Float
from sqlalchemy.orm import relationship
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.core.entities.model.model_datatype import DataTypeMatrix
from tvb.core.neotraits.db import from_ndarray
from tvb.datatypes.spectral import FourierSpectrum, WaveletCoefficients, CoherenceSpectrum, ComplexCoherenceSpectrum


class FourierSpectrumIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    segment_length = Column(Float, nullable=False)
    windowing_function = Column(String, nullable=True)
    frequency_step = Column(Float, nullable=False)
    max_frequency = Column(Float, nullable=False)

    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid), nullable=not FourierSpectrum.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    def fill_from_has_traits(self, datatype):
        # type: (FourierSpectrum)  -> None
        super(FourierSpectrumIndex, self).fill_from_has_traits(datatype)
        self.segment_length = datatype.segment_length
        self.windowing_function = datatype.windowing_function
        self.frequency_step = datatype.freq_step
        self.max_frequency = datatype.max_freq
        self.fk_source_gid = datatype.source.gid.hex

    def get_extra_info(self):
        labels_dict = {}
        labels_dict["labels_ordering"] = self.source.labels_ordering
        labels_dict["labels_dimensions"] = self.source.labels_dimensions
        return labels_dict


class WaveletCoefficientsIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid),
                           nullable=not WaveletCoefficients.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    mother = Column(String, nullable=False)
    normalisation = Column(String, nullable=False)
    q_ratio = Column(Float, nullable=False)
    sample_period = Column(Float, nullable=False)
    number_of_scales = Column(Integer, nullable=False)

    frequencies_min = Column(Float)
    frequencies_max = Column(Float)

    def fill_from_has_traits(self, datatype):
        # type: (WaveletCoefficients)  -> None
        super(WaveletCoefficientsIndex, self).fill_from_has_traits(datatype)
        self.mother = datatype.mother
        self.normalisation = datatype.normalisation
        self.q_ratio = datatype.q_ratio
        self.sample_period = datatype.sample_period
        self.number_of_scales = datatype.frequencies.shape[0]
        self.frequencies_min, self.frequencies_max, _ = from_ndarray(datatype.frequencies)
        self.fk_source_gid = datatype.source.gid.hex

    def get_extra_info(self):
        labels_dict = {}
        labels_dict["labels_ordering"] = self.source.labels_ordering
        labels_dict["labels_dimensions"] = self.source.labels_dimensions
        return labels_dict


class CoherenceSpectrumIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid), nullable=not CoherenceSpectrum.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    nfft = Column(Integer, nullable=False)
    frequencies_min = Column(Float)
    frequencies_max = Column(Float)

    def fill_from_has_traits(self, datatype):
        # type: (CoherenceSpectrum)  -> None
        super(CoherenceSpectrumIndex, self).fill_from_has_traits(datatype)
        self.nfft = datatype.nfft
        self.frequencies_min, self.frequencies_max, _ = from_ndarray(datatype.frequency)
        self.fk_source_gid = datatype.source.gid.hex

    def get_extra_info(self):
        labels_dict = {}
        labels_dict["labels_ordering"] = self.source.labels_ordering
        labels_dict["labels_dimensions"] = self.source.labels_dimensions
        return labels_dict


class ComplexCoherenceSpectrumIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid),
                           nullable=not ComplexCoherenceSpectrum.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    epoch_length = Column(Float, nullable=False)
    segment_length = Column(Float, nullable=False)
    windowing_function = Column(String, nullable=False)
    frequency_step = Column(Float, nullable=False)
    max_frequency = Column(Float, nullable=False)

    def fill_from_has_traits(self, datatype):
        # type: (ComplexCoherenceSpectrum)  -> None
        super(ComplexCoherenceSpectrumIndex, self).fill_from_has_traits(datatype)
        self.epoch_length = datatype.epoch_length
        self.segment_length = datatype.segment_length
        self.windowing_function = datatype.windowing_function
        self.frequency_step = datatype.freq_step
        self.max_frequency = datatype.max_freq
        self.fk_source_gid = datatype.source.gid.hex

    def get_extra_info(self):
        labels_dict = {}
        labels_dict["labels_ordering"] = self.source.labels_ordering
        labels_dict["labels_dimensions"] = self.source.labels_dimensions
        return labels_dict



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
"""

from tvb.adapters.datatypes.db.local_connectivity import LocalConnectivityIndex
from tvb.adapters.datatypes.db.surface import SurfaceIndex
from tvb.adapters.simulator.equation_forms import GaussianEquationForm, get_form_for_equation
from tvb.adapters.simulator.subforms_mapping import GAUSSIAN_EQUATION, get_ui_name_to_equation_dict, \
    DOUBLE_GAUSSIAN_EQUATION, SIGMOID_EQUATION
from tvb.basic.neotraits.api import Attr
from tvb.core.adapters.abcadapter import ABCAdapterForm, ABCAdapter
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import FormField, SelectField, TraitDataTypeSelectField, FloatField, StrField
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr, Str
from tvb.datatypes.local_connectivity import LocalConnectivity
from tvb.datatypes.surfaces import Surface, CORTICAL


class LocalConnectivitySelectorForm(ABCAdapterForm):

    def __init__(self):
        super(LocalConnectivitySelectorForm, self).__init__()
        traited_attr = Attr(self.get_required_datatype(), label='Load Local Connectivity', required=False)
        self.existentEntitiesSelect = TraitDataTypeSelectField(traited_attr, name='existentEntitiesSelect')

    @staticmethod
    def get_required_datatype():
        return LocalConnectivityIndex

    @staticmethod
    def get_input_name():
        pass

    @staticmethod
    def get_filters():
        return None

    def get_rendering_dict(self):
        return {'adapter_form': self, 'legend': 'Selected entity'}


class LocalConnectivityCreatorModel(ViewModel, LocalConnectivity):
    surface = DataTypeGidAttr(
        linked_datatype=Surface,
        label=LocalConnectivity.surface.label
    )

    display_name = Str(
        label='Display name',
        required=False
    )


class LocalConnectivityCreatorForm(ABCAdapterForm):
    NAME_EQUATION_PARAMS_DIV = 'spatial_params'
    possible_equations = {GAUSSIAN_EQUATION: get_ui_name_to_equation_dict().get(GAUSSIAN_EQUATION),
                          DOUBLE_GAUSSIAN_EQUATION: get_ui_name_to_equation_dict().get(DOUBLE_GAUSSIAN_EQUATION),
                          SIGMOID_EQUATION: get_ui_name_to_equation_dict().get(SIGMOID_EQUATION)}

    def __init__(self):
        super(LocalConnectivityCreatorForm, self).__init__()
        self.surface = TraitDataTypeSelectField(LocalConnectivityCreatorModel.surface, name=self.get_input_name(),
                                                conditions=self.get_filters())
        self.spatial = SelectField(LocalConnectivityCreatorModel.equation, name='spatial',
                                   choices=self.possible_equations,
                                   display_none_choice=False, subform=GaussianEquationForm)
        self.cutoff = FloatField(LocalConnectivityCreatorModel.cutoff)
        self.display_name = StrField(LocalConnectivityCreatorModel.display_name, name='display_name')

    @staticmethod
    def get_view_model():
        return LocalConnectivityCreatorModel

    @staticmethod
    def get_required_datatype():
        return SurfaceIndex

    @staticmethod
    def get_input_name():
        return 'surface'

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + '.surface_type'], operations=["=="],
                           values=[CORTICAL])

    def fill_from_trait(self, trait):
        # type: (LocalConnectivityCreatorModel) -> None
        self.surface.data = trait.surface.hex
        self.cutoff.data = trait.cutoff
        self.display_name.data = trait.display_name
        if trait.equation:
            lc_equation = trait.equation
        else:
            lc_equation = LocalConnectivity.equation.default
        self.spatial.data = type(lc_equation)
        self.spatial.subform_field = FormField(get_form_for_equation(type(lc_equation)), self.NAME_EQUATION_PARAMS_DIV)
        self.spatial.subform_field.form.fill_from_trait(lc_equation)

    def get_rendering_dict(self):
        return {'adapter_form': self, 'next_action': 'form_spatial_local_connectivity_data',
                'equation_params_div': self.NAME_EQUATION_PARAMS_DIV, 'legend': 'Local connectivity parameters'}


class LocalConnectivityCreator(ABCAdapter):
    """
    The purpose of this adapter is create a LocalConnectivity.
    """
    KEY_SURFACE = 'surface'
    KEY_EQUATION = 'equation'
    KEY_CUTOFF = 'cutoff'
    KEY_DISPLAY_NAME = 'display_name'

    def get_form_class(self):
        return LocalConnectivityCreatorForm

    def get_output(self):
        """
        Describes the outputs of the launch method.
        """
        return [LocalConnectivityIndex]

    def launch(self, view_model):
        # type: (LocalConnectivityCreatorModel) -> [LocalConnectivityIndex]
        """
        Used for creating a `LocalConnectivity`
        """
        local_connectivity = LocalConnectivity()
        local_connectivity.cutoff = view_model.cutoff
        if not self.surface_index:
            self.surface_index = self.load_entity_by_gid(view_model.surface)
        surface = h5.load_from_index(self.surface_index)
        local_connectivity.surface = surface
        local_connectivity.equation = view_model.equation
        local_connectivity.compute_sparse_matrix()
        self.generic_attributes.user_tag_1 = view_model.display_name

        return h5.store_complete(local_connectivity, self.storage_path)

    def get_required_disk_size(self, view_model):
        # type: (LocalConnectivityCreatorModel) -> int
        """
        Returns the required disk size to be able to run the adapter. (in kB)
        """
        if view_model.surface:
            self.surface_index = self.load_entity_by_gid(view_model.surface)
            points_no = view_model.cutoff / self.surface_index.edge_mean_length
            disk_size_b = self.surface_index.number_of_vertices * points_no * points_no * 8
            return self.array_size2kb(disk_size_b)
        return 0

    def get_required_memory_size(self, view_model):
        # type: (LocalConnectivityCreatorModel) -> int
        """
        Return the required memory to run this algorithm.
        """
        return self.get_required_disk_size(view_model)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from tvb.adapters.simulator.form_with_ranges import FormWithRanges
from tvb.core.neotraits.forms import Form, ArrayField, MultiSelectField
from tvb.simulator.models import ModelsEnum


def get_model_to_form_dict():
    model_class_to_form = {
        ModelsEnum.GENERIC_2D_OSCILLATOR.get_class(): Generic2dOscillatorModelForm,
        ModelsEnum.KURAMOTO.get_class(): KuramotoModelForm,
        ModelsEnum.SUP_HOPF.get_class(): SupHopfModelForm,
        ModelsEnum.HOPFIELD.get_class(): HopfieldModelForm,
        ModelsEnum.EPILEPTOR.get_class(): EpileptorModelForm,
        ModelsEnum.EPILEPTOR_2D.get_class(): Epileptor2DModelForm,
        ModelsEnum.EPILEPTOR_CODIM_3.get_class(): EpileptorCodim3ModelForm,
        ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class(): EpileptorCodim3SlowModModelForm,
        ModelsEnum.EPILEPTOR_RS.get_class(): EpileptorRestingStateModelForm,
        ModelsEnum.JANSEN_RIT.get_class(): JansenRitModelForm,
        ModelsEnum.ZETTERBERG_JANSEN.get_class(): ZetterbergJansenModelForm,
        ModelsEnum.REDUCED_WONG_WANG.get_class(): ReducedWongWangModelForm,
        ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class(): ReducedWongWangExcInhModelForm,
        ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class(): ReducedSetFitzHughNagumoModelForm,
        ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class(): ReducedSetHindmarshRoseModelForm,
        ModelsEnum.ZERLAUT_FIRST_ORDER.get_class(): ZerlautAdaptationFirstOrderModelForm,
        ModelsEnum.ZERLAUT_SECOND_ORDER.get_class(): ZerlautAdaptationSecondOrderModelForm,
        ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class(): MontbrioPazoRoxinModelForm,
        ModelsEnum.COOMBES_BYRNE.get_class(): CoombesByrneModelForm,
        ModelsEnum.COOMBES_BYRNE_2D.get_class(): CoombesByrne2DModelForm,
        ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class(): GastSchmidtKnoscheSDModelForm,
        ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class(): GastSchmidtKnoscheSFModelForm,
        ModelsEnum.DUMONT_GUTKIN.get_class(): DumontGutkinModelForm,
        ModelsEnum.LINEAR.get_class(): LinearModelForm,
        ModelsEnum.WILSON_COWAN.get_class(): WilsonCowanModelForm,
        ModelsEnum.LARTER_BREAKSPEAR.get_class(): LarterBreakspearModelForm
    }

    return model_class_to_form


def get_ui_name_to_model():
    ui_name_to_model = {
        'Generic 2d Oscillator': ModelsEnum.GENERIC_2D_OSCILLATOR.get_class(),
        'Kuramoto Oscillator': ModelsEnum.KURAMOTO.get_class(),
        'supHopf': ModelsEnum.SUP_HOPF.get_class(),
        'Hopfield': ModelsEnum.HOPFIELD.get_class(),
        'Epileptor': ModelsEnum.EPILEPTOR.get_class(),
        'Epileptor2D': ModelsEnum.EPILEPTOR_2D.get_class(),
        'Epileptor codim 3': ModelsEnum.EPILEPTOR_CODIM_3.get_class(),
        'Epileptor codim 3 ultra-slow modulations': ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class(),
        'Epileptor Resting State': ModelsEnum.EPILEPTOR_RS.get_class(),
        'Jansen-Rit': ModelsEnum.JANSEN_RIT.get_class(),
        'Zetterberg-Jansen': ModelsEnum.ZETTERBERG_JANSEN.get_class(),
        'Reduced Wong-Wang': ModelsEnum.REDUCED_WONG_WANG.get_class(),
        'Reduced Wong-Wang with Excitatory and Inhibitory Coupled Populations': ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class(),
        'Stefanescu-Jirsa 2D': ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class(),
        'Stefanescu-Jirsa 3D': ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class(),
        'Zerlaut adaptation first order': ModelsEnum.ZERLAUT_FIRST_ORDER.get_class(),
        'Zerlaut adaptation second order': ModelsEnum.ZERLAUT_SECOND_ORDER.get_class(),
        'Montbrio Pazo Roxin': ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class(),
        'Coombes Byrne': ModelsEnum.COOMBES_BYRNE.get_class(),
        'Coombes Byrne 2D': ModelsEnum.COOMBES_BYRNE_2D.get_class(),
        'Gast Schmidt Knosche_SD': ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class(),
        'Gast Schmidt Knosche_SF': ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class(),
        'Dumont Gutkin': ModelsEnum.DUMONT_GUTKIN.get_class(),
        'Linear model': ModelsEnum.LINEAR.get_class(),
        'Wilson-Cowan': ModelsEnum.WILSON_COWAN.get_class(),
        'Larter-Breakspear': ModelsEnum.LARTER_BREAKSPEAR.get_class()
    }
    return ui_name_to_model


def get_form_for_model(model_class):
    return get_model_to_form_dict().get(model_class)


class StateVariableRangesForm(Form):

    def __init__(self):
        super(StateVariableRangesForm, self).__init__()


class Generic2dOscillatorModelForm(FormWithRanges):

    def __init__(self):
        super(Generic2dOscillatorModelForm, self).__init__()
        self.tau = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().tau)
        self.I = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().I)
        self.a = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().a)
        self.b = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().b)
        self.c = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().c)
        self.d = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().d)
        self.e = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().e)
        self.f = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().f)
        self.g = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().g)
        self.alpha = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().alpha)
        self.beta = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().beta)
        self.gamma = ArrayField(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().gamma)
        self.variables_of_interest = MultiSelectField(
            ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'a', 'b', 'c', 'I', 'd', 'e', 'f', 'g', 'alpha', 'beta', 'gamma']


class KuramotoModelForm(FormWithRanges):

    def __init__(self):
        super(KuramotoModelForm, self).__init__()
        self.omega = ArrayField(ModelsEnum.KURAMOTO.get_class().omega)
        self.variables_of_interest = MultiSelectField(ModelsEnum.KURAMOTO.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['omega']


class SupHopfModelForm(FormWithRanges):

    def __init__(self):
        super(SupHopfModelForm, self).__init__()
        self.a = ArrayField(ModelsEnum.SUP_HOPF.get_class().a)
        self.omega = ArrayField(ModelsEnum.SUP_HOPF.get_class().omega)
        self.variables_of_interest = MultiSelectField(ModelsEnum.SUP_HOPF.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['a', 'omega']


class HopfieldModelForm(FormWithRanges):

    def __init__(self):
        super(HopfieldModelForm, self).__init__()
        self.taux = ArrayField(ModelsEnum.HOPFIELD.get_class().taux)
        self.tauT = ArrayField(ModelsEnum.HOPFIELD.get_class().tauT)
        self.dynamic = ArrayField(ModelsEnum.HOPFIELD.get_class().dynamic)
        self.variables_of_interest = MultiSelectField(ModelsEnum.HOPFIELD.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['taux', 'tauT', 'dynamic']


class EpileptorModelForm(FormWithRanges):

    def __init__(self):
        super(EpileptorModelForm, self).__init__()
        self.a = ArrayField(ModelsEnum.EPILEPTOR.get_class().a)
        self.b = ArrayField(ModelsEnum.EPILEPTOR.get_class().b)
        self.c = ArrayField(ModelsEnum.EPILEPTOR.get_class().c)
        self.d = ArrayField(ModelsEnum.EPILEPTOR.get_class().d)
        self.r = ArrayField(ModelsEnum.EPILEPTOR.get_class().r)
        self.s = ArrayField(ModelsEnum.EPILEPTOR.get_class().s)
        self.x0 = ArrayField(ModelsEnum.EPILEPTOR.get_class().x0)
        self.Iext = ArrayField(ModelsEnum.EPILEPTOR.get_class().Iext)
        self.slope = ArrayField(ModelsEnum.EPILEPTOR.get_class().slope)
        self.Iext2 = ArrayField(ModelsEnum.EPILEPTOR.get_class().Iext2)
        self.tau = ArrayField(ModelsEnum.EPILEPTOR.get_class().tau)
        self.aa = ArrayField(ModelsEnum.EPILEPTOR.get_class().aa)
        self.bb = ArrayField(ModelsEnum.EPILEPTOR.get_class().bb)
        self.Kvf = ArrayField(ModelsEnum.EPILEPTOR.get_class().Kvf)
        self.Kf = ArrayField(ModelsEnum.EPILEPTOR.get_class().Kf)
        self.Ks = ArrayField(ModelsEnum.EPILEPTOR.get_class().Ks)
        self.tt = ArrayField(ModelsEnum.EPILEPTOR.get_class().tt)
        self.modification = ArrayField(ModelsEnum.EPILEPTOR.get_class().modification)
        self.variables_of_interest = MultiSelectField(ModelsEnum.EPILEPTOR.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ["Iext", "Iext2", "r", "x0", "slope"]


class Epileptor2DModelForm(FormWithRanges):

    def __init__(self):
        super(Epileptor2DModelForm, self).__init__()
        self.a = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().a)
        self.b = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().b)
        self.c = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().c)
        self.d = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().d)
        self.r = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().r)
        self.x0 = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().x0)
        self.Iext = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().Iext)
        self.slope = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().slope)
        self.Kvf = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().Kvf)
        self.Ks = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().Ks)
        self.tt = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().tt)
        self.modification = ArrayField(ModelsEnum.EPILEPTOR_2D.get_class().modification)
        self.variables_of_interest = MultiSelectField(ModelsEnum.EPILEPTOR_2D.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ["r", "Iext", "x0"]


class EpileptorCodim3ModelForm(FormWithRanges):

    def __init__(self):
        super(EpileptorCodim3ModelForm, self).__init__()
        self.mu1_start = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu1_start)
        self.mu2_start = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu2_start)
        self.nu_start = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().nu_start)
        self.mu1_stop = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu1_stop)
        self.mu2_stop = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu2_stop)
        self.nu_stop = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().nu_stop)
        self.b = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().b)
        self.R = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().R)
        self.c = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().c)
        self.dstar = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().dstar)
        self.Ks = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().Ks)
        self.N = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().N)
        self.modification = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().modification)
        self.variables_of_interest = MultiSelectField(ModelsEnum.EPILEPTOR_CODIM_3.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['mu1_start', 'mu2_start', 'nu_start', 'mu1_stop', 'mu2_stop', 'nu_stop', 'b', 'R', 'c', 'dstar', 'N',
                'Ks']


class EpileptorCodim3SlowModModelForm(FormWithRanges):

    def __init__(self):
        super(EpileptorCodim3SlowModModelForm, self).__init__()
        self.mu1_Ain = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Ain)
        self.mu2_Ain = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Ain)
        self.nu_Ain = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Ain)
        self.mu1_Bin = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Bin)
        self.mu2_Bin = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Bin)
        self.nu_Bin = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Bin)
        self.mu1_Aend = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Aend)
        self.mu2_Aend = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Aend)
        self.nu_Aend = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Aend)
        self.mu1_Bend = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Bend)
        self.mu2_Bend = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Bend)
        self.nu_Bend = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Bend)
        self.b = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().b)
        self.R = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().R)
        self.c = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().c)
        self.cA = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().cA)
        self.cB = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().cB)
        self.dstar = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().dstar)
        self.Ks = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().Ks)
        self.N = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().N)
        self.modification = ArrayField(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().modification)
        self.variables_of_interest = MultiSelectField(
            ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['mu1_Ain', 'mu2_Ain', 'nu_Ain', 'mu1_Bin', 'mu2_Bin', 'nu_Bin', 'mu1_Aend', 'mu2_Aend', 'nu_Aend',
                'mu1_Bend', 'mu2_Bend', 'nu_Bend', 'b', 'R', 'c', 'dstar', 'N']


class EpileptorRestingStateModelForm(FormWithRanges):

    def __init__(self):
        super(EpileptorRestingStateModelForm, self).__init__()
        self.a = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().a)
        self.b = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().b)
        self.c = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().c)
        self.d = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().d)
        self.r = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().r)
        self.s = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().s)
        self.x0 = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().x0)
        self.Iext = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().Iext)
        self.slope = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().slope)
        self.Iext2 = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().Iext2)
        self.tau = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().tau)
        self.aa = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().aa)
        self.bb = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().bb)
        self.Kvf = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().Kvf)
        self.Kf = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().Kf)
        self.Ks = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().Ks)
        self.tt = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().tt)
        self.tau_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().tau_rs)
        self.I_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().I_rs)
        self.a_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().a_rs)
        self.b_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().b_rs)
        self.d_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().d_rs)
        self.e_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().e_rs)
        self.f_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().f_rs)
        self.alpha_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().alpha_rs)
        self.beta_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().beta_rs)
        self.K_rs = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().K_rs)
        self.p = ArrayField(ModelsEnum.EPILEPTOR_RS.get_class().p)
        self.variables_of_interest = MultiSelectField(ModelsEnum.EPILEPTOR_RS.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['Iext', 'Iext2', 'r', 'x0', 'slope', 'tau_rs', 'a_rs', 'b_rs', 'I_rs', 'd_rs', 'e_rs', 'f_rs',
                'alpha_rs', 'beta_rs', 'gamma_rs']


class JansenRitModelForm(FormWithRanges):

    def __init__(self):
        super(JansenRitModelForm, self).__init__()
        self.A = ArrayField(ModelsEnum.JANSEN_RIT.get_class().A)
        self.B = ArrayField(ModelsEnum.JANSEN_RIT.get_class().B)
        self.a = ArrayField(ModelsEnum.JANSEN_RIT.get_class().a)
        self.b = ArrayField(ModelsEnum.JANSEN_RIT.get_class().b)
        self.v0 = ArrayField(ModelsEnum.JANSEN_RIT.get_class().v0)
        self.nu_max = ArrayField(ModelsEnum.JANSEN_RIT.get_class().nu_max)
        self.r = ArrayField(ModelsEnum.JANSEN_RIT.get_class().r)
        self.J = ArrayField(ModelsEnum.JANSEN_RIT.get_class().J)
        self.a_1 = ArrayField(ModelsEnum.JANSEN_RIT.get_class().a_1)
        self.a_2 = ArrayField(ModelsEnum.JANSEN_RIT.get_class().a_2)
        self.a_3 = ArrayField(ModelsEnum.JANSEN_RIT.get_class().a_3)
        self.a_4 = ArrayField(ModelsEnum.JANSEN_RIT.get_class().a_4)
        self.p_min = ArrayField(ModelsEnum.JANSEN_RIT.get_class().p_min)
        self.p_max = ArrayField(ModelsEnum.JANSEN_RIT.get_class().p_max)
        self.mu = ArrayField(ModelsEnum.JANSEN_RIT.get_class().mu)
        self.variables_of_interest = MultiSelectField(ModelsEnum.JANSEN_RIT.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['A', 'B', 'a', 'b', 'v0', 'nu_max', 'r', 'J', 'a_1', 'a_2', 'a_3', 'a_4', 'p_min', 'p_max', 'mu']


class ZetterbergJansenModelForm(FormWithRanges):

    def __init__(self):
        super(ZetterbergJansenModelForm, self).__init__()
        self.He = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().He)
        self.Hi = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().Hi)
        self.ke = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().ke)
        self.ki = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().ki)
        self.e0 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().e0)
        self.rho_2 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().rho_2)
        self.rho_1 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().rho_1)
        self.gamma_1 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_1)
        self.gamma_2 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_2)
        self.gamma_3 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_3)
        self.gamma_4 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_4)
        self.gamma_5 = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_5)
        self.gamma_1T = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_1T)
        self.gamma_2T = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_2T)
        self.gamma_3T = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_3T)
        self.P = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().P)
        self.U = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().U)
        self.Q = ArrayField(ModelsEnum.ZETTERBERG_JANSEN.get_class().Q)
        self.variables_of_interest = MultiSelectField(ModelsEnum.ZETTERBERG_JANSEN.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['He', 'Hi', 'ke', 'ki', 'e0', 'rho_2', 'rho_1', 'gamma_1', 'gamma_2', 'gamma_3', 'gamma_4', 'gamma_5',
                'P', 'U', 'Q']


class ReducedWongWangModelForm(FormWithRanges):

    def __init__(self):
        super(ReducedWongWangModelForm, self).__init__()
        self.a = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().a)
        self.b = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().b)
        self.d = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().d)
        self.gamma = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().gamma)
        self.tau_s = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().tau_s)
        self.w = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().w)
        self.J_N = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().J_N)
        self.I_o = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().I_o)
        self.sigma_noise = ArrayField(ModelsEnum.REDUCED_WONG_WANG.get_class().sigma_noise)
        self.variables_of_interest = MultiSelectField(ModelsEnum.REDUCED_WONG_WANG.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['a', 'b', 'd', 'gamma', 'tau_s', 'w', 'J_N', 'I_o']


class ReducedWongWangExcInhModelForm(FormWithRanges):

    def __init__(self):
        super(ReducedWongWangExcInhModelForm, self).__init__()
        self.a_e = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().a_e)
        self.b_e = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().b_e)
        self.d_e = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().d_e)
        self.gamma_e = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().gamma_e)
        self.tau_e = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().tau_e)
        self.w_p = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().w_p)
        self.J_N = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().J_N)
        self.W_e = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().W_e)
        self.a_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().a_i)
        self.b_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().b_i)
        self.d_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().d_i)
        self.gamma_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().gamma_i)
        self.tau_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().tau_i)
        self.J_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().J_i)
        self.W_i = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().W_i)
        self.I_o = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().I_o)
        self.G = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().G)
        self.lamda = ArrayField(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().lamda)
        self.variables_of_interest = MultiSelectField(
            ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['a_e', 'b_e', 'd_e', 'gamma_e', 'tau_e', 'W_e', 'w_p', 'J_N', 'a_i', 'b_i', 'd_i', 'gamma_i', 'tau_i',
                'W_i', 'J_i', 'I_o', 'G', 'lamda']


class ReducedSetFitzHughNagumoModelForm(FormWithRanges):

    def __init__(self):
        super(ReducedSetFitzHughNagumoModelForm, self).__init__()
        self.tau = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().tau)
        self.a = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().a)
        self.b = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().b)
        self.K11 = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().K11)
        self.K12 = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().K12)
        self.K21 = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().K21)
        self.sigma = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().sigma)
        self.mu = ArrayField(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().mu)
        self.variables_of_interest = MultiSelectField(
            ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'a', 'b', 'K11', 'K12', 'K21', 'sigma', 'mu']


class ReducedSetHindmarshRoseModelForm(FormWithRanges):

    def __init__(self):
        super(ReducedSetHindmarshRoseModelForm, self).__init__()
        self.r = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().r)
        self.a = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().a)
        self.b = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().b)
        self.c = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().c)
        self.d = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().d)
        self.s = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().s)
        self.xo = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().xo)
        self.K11 = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().K11)
        self.K12 = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().K12)
        self.K21 = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().K21)
        self.sigma = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().sigma)
        self.mu = ArrayField(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().mu)
        self.variables_of_interest = MultiSelectField(
            ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['r', 'a', 'b', 'c', 'd', 's', 'xo', 'K11', 'K12', 'K21', 'sigma', 'mu']


class ZerlautAdaptationFirstOrderModelForm(FormWithRanges):

    def __init__(self):
        super(ZerlautAdaptationFirstOrderModelForm, self).__init__()
        self.g_L = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().g_L)
        self.E_L_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_L_e)
        self.E_L_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_L_i)
        self.C_m = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().C_m)
        self.b_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().b_e)
        self.b_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().b_i)
        self.a_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().a_e)
        self.a_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().a_i)
        self.tau_w_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_w_e)
        self.tau_w_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_w_i)
        self.E_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_e)
        self.E_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_i)
        self.Q_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().Q_e)
        self.Q_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().Q_i)
        self.tau_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_e)
        self.tau_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_i)
        self.N_tot = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().N_tot)
        self.p_connect = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().p_connect)
        self.g = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().g)
        self.K_ext_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().K_ext_e)
        self.K_ext_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().K_ext_i)
        self.T = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().T)
        self.P_e = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().P_e)
        self.P_i = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().P_i)
        self.external_input_ex_ex = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_ex_ex)
        self.external_input_ex_in = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_ex_in)
        self.external_input_in_ex = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_in_ex)
        self.external_input_in_in = ArrayField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_in_in)
        self.variables_of_interest = MultiSelectField(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['g_L', 'E_L_e', 'E_L_i', 'C_m', 'b_e', 'b_i', 'a_e', 'a_i', 'tau_w_e', 'tau_w_i', 'E_e', 'E_i', 'Q_e',
                'Q_i', 'tau_e', 'tau_i', 'N_tot', 'p_connect', 'g', 'K_ext_e', 'K_ext_i', 'T', 'external_input_ex_ex',
                'external_input_ex_in', 'external_input_in_ex', 'external_input_in_in']


class ZerlautAdaptationSecondOrderModelForm(ZerlautAdaptationFirstOrderModelForm):

    def __init__(self):
        super(ZerlautAdaptationSecondOrderModelForm, self).__init__()
        self.variables_of_interest = MultiSelectField(ModelsEnum.ZERLAUT_SECOND_ORDER.get_class().variables_of_interest)

class MontbrioPazoRoxinModelForm(FormWithRanges):

    def __init__(self):
        super(MontbrioPazoRoxinModelForm, self).__init__()
        self.tau = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().tau)
        self.I = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().I)
        self.Delta = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().Delta)
        self.J = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().J)
        self.eta = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().eta)
        self.Gamma = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().Gamma)
        self.cr = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().cr)
        self.cv = ArrayField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().cv)
        self.variables_of_interest = MultiSelectField(ModelsEnum.MONTBRIO_PAZO_ROXIN.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'I', 'Delta', 'J', 'eta', 'Gamma', 'cr', 'cv']

class CoombesByrneModelForm(FormWithRanges):

    def __init__(self):
        super(CoombesByrneModelForm, self).__init__()
        self.Delta = ArrayField(ModelsEnum.COOMBES_BYRNE.get_class().Delta)
        self.alpha = ArrayField(ModelsEnum.COOMBES_BYRNE.get_class().alpha)
        self.v_syn = ArrayField(ModelsEnum.COOMBES_BYRNE.get_class().v_syn)
        self.k = ArrayField(ModelsEnum.COOMBES_BYRNE.get_class().k)
        self.eta = ArrayField(ModelsEnum.COOMBES_BYRNE.get_class().eta)
        self.variables_of_interest = MultiSelectField(ModelsEnum.COOMBES_BYRNE.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['Delta', 'alpha', 'v_syn', 'k', 'eta']

class CoombesByrne2DModelForm(FormWithRanges):

    def __init__(self):
        super(CoombesByrne2DModelForm, self).__init__()
        self.Delta = ArrayField(ModelsEnum.COOMBES_BYRNE_2D.get_class().Delta)
        self.v_syn = ArrayField(ModelsEnum.COOMBES_BYRNE_2D.get_class().v_syn)
        self.k = ArrayField(ModelsEnum.COOMBES_BYRNE_2D.get_class().k)
        self.eta = ArrayField(ModelsEnum.COOMBES_BYRNE_2D.get_class().eta)
        self.variables_of_interest = MultiSelectField(ModelsEnum.COOMBES_BYRNE_2D.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['Delta', 'v_syn', 'k', 'eta']

class GastSchmidtKnoscheSDModelForm(FormWithRanges):

    def __init__(self):
        super(GastSchmidtKnoscheSDModelForm, self).__init__()
        self.tau = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().tau)
        self.tau_A = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().tau_A)
        self.alpha = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().alpha)
        self.I = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().I)
        self.Delta = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().Delta)
        self.J = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().J)
        self.eta = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().eta)
        self.cr = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().cr)
        self.cv = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().cv)
        self.variables_of_interest = MultiSelectField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'tau_A', 'alpha', 'I', 'Delta', 'J', 'eta', 'cr', 'cv']

class GastSchmidtKnoscheSFModelForm(FormWithRanges):

    def __init__(self):
        super(GastSchmidtKnoscheSFModelForm, self).__init__()
        self.tau = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().tau)
        self.tau_A = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().tau_A)
        self.alpha = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().alpha)
        self.I = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().I)
        self.Delta = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().Delta)
        self.J = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().J)
        self.eta = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().eta)
        self.cr = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().cr)
        self.cv = ArrayField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().cv)
        self.variables_of_interest = MultiSelectField(ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'tau_A', 'alpha', 'I', 'Delta', 'J', 'eta', 'cr', 'cv']

class DumontGutkinModelForm(FormWithRanges):

    def __init__(self):
        super(DumontGutkinModelForm, self).__init__()
        self.I_e = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().I_e)
        self.Delta_e = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().Delta_e)
        self.eta_e = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().eta_e)
        self.tau_e = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().tau_e)
        self.I_i = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().I_i)
        self.Delta_i = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().Delta_i)
        self.eta_i = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().eta_i)
        self.tau_i = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().tau_i)
        self.tau_s = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().tau_s)
        self.J_ee = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().J_ee)
        self.J_ei = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().J_ei)
        self.J_ie = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().J_ie)
        self.J_ii = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().J_ii)
        self.Gamma = ArrayField(ModelsEnum.DUMONT_GUTKIN.get_class().Gamma)
        self.variables_of_interest = MultiSelectField(ModelsEnum.DUMONT_GUTKIN.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['I_e', 'Delta_e', 'eta_e', 'tau_e', 'I_i', 'Delta_i', 'eta_i', 'tau_i', 'tau_s',
                'J_ee', 'J_ei', 'J_ie', 'J_ii', 'Gamma']


class LinearModelForm(FormWithRanges):

    def __init__(self):
        super(LinearModelForm, self).__init__()
        self.gamma = ArrayField(ModelsEnum.LINEAR.get_class().gamma)
        self.variables_of_interest = MultiSelectField(ModelsEnum.LINEAR.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['gamma']


class WilsonCowanModelForm(FormWithRanges):

    def __init__(self):
        super(WilsonCowanModelForm, self).__init__()
        self.c_ee = ArrayField(ModelsEnum.WILSON_COWAN.get_class().c_ee)
        self.c_ie = ArrayField(ModelsEnum.WILSON_COWAN.get_class().c_ie)
        self.c_ei = ArrayField(ModelsEnum.WILSON_COWAN.get_class().c_ei)
        self.c_ii = ArrayField(ModelsEnum.WILSON_COWAN.get_class().c_ii)
        self.tau_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().tau_e)
        self.tau_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().tau_i)
        self.a_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().a_e)
        self.b_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().b_e)
        self.c_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().c_e)
        self.theta_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().theta_e)
        self.a_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().a_i)
        self.b_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().b_i)
        self.theta_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().theta_i)
        self.c_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().c_i)
        self.r_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().r_e)
        self.r_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().r_i)
        self.k_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().k_e)
        self.k_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().k_i)
        self.P = ArrayField(ModelsEnum.WILSON_COWAN.get_class().P)
        self.Q = ArrayField(ModelsEnum.WILSON_COWAN.get_class().Q)
        self.alpha_e = ArrayField(ModelsEnum.WILSON_COWAN.get_class().alpha_e)
        self.alpha_i = ArrayField(ModelsEnum.WILSON_COWAN.get_class().alpha_i)
        self.variables_of_interest = MultiSelectField(ModelsEnum.WILSON_COWAN.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['c_ee', 'c_ei', 'c_ie', 'c_ii', 'tau_e', 'tau_i', 'a_e', 'b_e', 'c_e', 'a_i', 'b_i', 'c_i', 'r_e',
                'r_i', 'k_e', 'k_i', 'P', 'Q', 'theta_e', 'theta_i', 'alpha_e', 'alpha_i']


class LarterBreakspearModelForm(FormWithRanges):

    def __init__(self):
        super(LarterBreakspearModelForm, self).__init__()
        self.gCa = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().gCa)
        self.gK = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().gK)
        self.gL = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().gL)
        self.phi = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().phi)
        self.gNa = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().gNa)
        self.TK = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().TK)
        self.TCa = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().TCa)
        self.TNa = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().TNa)
        self.VCa = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().VCa)
        self.VK = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().VK)
        self.VL = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().VL)
        self.VNa = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().VNa)
        self.d_K = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_K)
        self.tau_K = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().tau_K)
        self.d_Na = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_Na)
        self.d_Ca = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_Ca)
        self.aei = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().aei)
        self.aie = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().aie)
        self.b = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().b)
        self.C = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().C)
        self.ane = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().ane)
        self.ani = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().ani)
        self.aee = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().aee)
        self.Iext = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().Iext)
        self.rNMDA = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().rNMDA)
        self.VT = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().VT)
        self.d_V = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_V)
        self.ZT = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().ZT)
        self.d_Z = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_Z)
        self.QV_max = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().QV_max)
        self.QZ_max = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().QZ_max)
        self.t_scale = ArrayField(ModelsEnum.LARTER_BREAKSPEAR.get_class().t_scale)
        self.variables_of_interest = MultiSelectField(ModelsEnum.LARTER_BREAKSPEAR.get_class().variables_of_interest)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['gCa', 'gK', 'gL', 'phi', 'gNa', 'TK', 'TCa', 'TNa', 'VCa', 'VK', 'VL', 'VNa', 'd_K', 'tau_K', 'd_Na',
                'd_Ca', 'aei', 'aie', 'b', 'C', 'ane', 'ani', 'aee', 'Iext', 'rNMDA', 'VT', 'd_V', 'ZT', 'd_Z',
                'QV_max', 'QZ_max', 't_scale']



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import xml.dom.minidom
from xml.dom.minidom import Node
from tvb.basic.logger.builder import get_logger
from tvb.core.adapters.exceptions import ParseException
from tvb.adapters.datatypes.db.annotation import AnnotationTerm


class XMLParser(object):
    """
    Parser for XML with Connectivity annotations
    """
    LOGGER = get_logger(__name__)

    ATTR_TVB_ID = "id"
    ATTR_TVB_LABEL = "label"
    PREFIX_TVB_ID = "http://thevirtualbrain.org#"
    PREFIX_LEFT = ["l", "L", "left", "Left"]
    PREFIX_RIGHT = ["r", "R", "right", "Right"]

    ATTR_ONTOLOGY_URI = "uri"
    ATTR_ONTOLOGY_RELATION = "relation"
    ATTR_ONTOLOGY_LABEL = "label"
    ATTR_ONTOLOGY_SYN = "synonym"
    ATTR_ONTOLOGY_DEFINITION = "definition"

    def __init__(self, xml_file, region_labels):
        self.doc_xml = xml.dom.minidom.parse(xml_file)
        self.connectivity_labels = [rl.lower() for rl in region_labels]
        self.last_id = 0

    def _generate_new_id(self):
        self.last_id += 1
        return self.last_id

    def _find_region_idxs(self, region_label):
        """
        Find IDX in Connectivity based on label read from XML
        :param region_label: Label ar read from XML (including PREFIX_TVB_ID)
        :return: (left_hemisphere, right_hemisphere_idx)
        :raise: ParseException in case of not match
        """
        short_name = region_label.lower().replace(self.PREFIX_TVB_ID, "")

        left_idx = [self.connectivity_labels.index(prefix + short_name) for prefix in self.PREFIX_LEFT if
                    prefix + short_name in self.connectivity_labels]
        right_idx = [self.connectivity_labels.index(prefix + short_name) for prefix in self.PREFIX_RIGHT if
                     prefix + short_name in self.connectivity_labels]

        if len(left_idx) > 0 and len(right_idx) > 0:
            return left_idx[0], right_idx[0]

        raise ParseException("Could not match regionID '%s' from XML with the chosen connectivity" % region_label)

    def _parse_ontology_children(self, parent_node, parent_term_id, region_left, region_right):
        """
        To be called recursively, for processing OntologyTerm XML tags.
        """
        result = []
        for ont_node in parent_node.childNodes:
            if ont_node.nodeType != Node.ELEMENT_NODE:
                continue

            ont_uri = ont_node.getAttribute(self.ATTR_ONTOLOGY_URI)
            ont_relation = ont_node.getAttribute(self.ATTR_ONTOLOGY_RELATION)
            ont_label = ont_node.getAttribute(self.ATTR_ONTOLOGY_LABEL)
            ont_syn = ont_node.getAttribute(self.ATTR_ONTOLOGY_SYN)
            ont_definition = ont_node.getAttribute(self.ATTR_ONTOLOGY_DEFINITION)

            # if len(ont_syn) > 0 and ont_syn[:len(ont_syn) / 2] == ont_syn[len(ont_syn) / 2 + 1:]:
            #     ont_syn = ont_syn[:len(ont_syn) / 2]

            try:
                # Check if it is a TVB region
                syn_left, syn_right = self._find_region_idxs(ont_uri)
                ont_term = AnnotationTerm(self._generate_new_id(), parent_term_id, region_left, region_right,
                                          ont_relation, ont_label, ont_definition, ont_syn, ont_uri,
                                          syn_left, syn_right)
            except ParseException:

                ont_term = AnnotationTerm(self._generate_new_id(), parent_term_id, region_left, region_right,
                                          ont_relation, ont_label, ont_definition, ont_syn, ont_uri)
            # Call recursively for all child nodes
            child_terms = self._parse_ontology_children(ont_node, ont_term.id, region_left, region_right)
            result.append(ont_term)
            result.extend(child_terms)
        return result

    def read_annotation_terms(self):
        """
        Process XML file.
        :return List of AnnotationTerm instances.
        """
        result = []
        for tvb_node in self.doc_xml.lastChild.childNodes:
            if tvb_node.nodeType != Node.ELEMENT_NODE:
                continue

            tvb_id = tvb_node.getAttribute(self.ATTR_TVB_ID)
            tvb_label = tvb_node.getAttribute(self.ATTR_TVB_LABEL)
            self.LOGGER.debug("Processing region %s:%s" % (tvb_id, tvb_label))
            region_left, region_right = self._find_region_idxs(tvb_id)
            self.LOGGER.debug("Matched %s == (%d, %d)" % (tvb_label, region_left, region_right))
            ont_terms = self._parse_ontology_children(tvb_node, -1, region_left, region_right)
            result.extend(ont_terms)
        return result



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
TVB global configurations are predefined/read from here, for working with Framework.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import os
import sys

from tvb.basic.config import stored
from tvb.basic.config.profile_settings import BaseSettingsProfile
from tvb.basic.config.settings import DBSettings
from tvb.basic.logger import builder
from tvb.basic.logger.builder import LoggerBuilder
from tvb.config.init.datatypes_registry import populate_datatypes_registry


class WebSettingsProfile(BaseSettingsProfile):
    """
    Setting for working with storage and web interface
    """
    LOGGER_CONFIG_FILE_NAME = "logger_config.conf"

    def initialize_profile(self, change_logger_in_dev=True):
        """
        Specific initialization when functioning with storage
        """
        super(WebSettingsProfile, self).initialize_profile()

        if change_logger_in_dev and not self.env.is_distribution():
            self.LOGGER_CONFIG_FILE_NAME = "dev_logger_config.conf"

        populate_datatypes_registry()
        builder.GLOBAL_LOGGER_BUILDER = LoggerBuilder('tvb.config.logger')

    def initialize_for_deployment(self):
        """
        Specific initialization for deployment and framework settings
        """
        super(WebSettingsProfile, self).initialize_for_deployment()

        inside_static_folder = os.path.join(self.EXTERNALS_FOLDER_PARENT, 'tvb')
        self.web.CHERRYPY_CONFIGURATION['/statichelp']['tools.staticdir.root'] = inside_static_folder

        # We want to disable warnings we get from sqlalchemy for traited attributes when we are in deployment mode.
        import warnings
        from sqlalchemy import exc as sa_exc

        warnings.simplefilter("ignore", category=sa_exc.SAWarning)


class CommandSettingsProfile(WebSettingsProfile):
    """
    Profile which allows you to work in tvb with storage enable, but in console mode.
    """


class TestSQLiteProfile(WebSettingsProfile):
    """
    Defines settings for running tests on an SQLite database.
    """

    # Use a different configuration file, to make it possible to run multiple instances in the same time
    TVB_CONFIG_FILE = os.path.expanduser(os.path.join("~", '.test.tvb.configuration'))

    DEFAULT_STORAGE = os.path.expanduser(os.path.join('~', 'TVB_TEST'))

    REVISION_NUMBER = 1
    CODE_CHECKED_TO_VERSION = sys.maxsize
    TRADE_CRASH_SAFETY_FOR_SPEED = True

    def __init__(self):
        super(TestSQLiteProfile, self).__init__()

        self.web.RENDER_HTML = False
        self.MAX_THREADS_NUMBER = self.manager.get_attribute(stored.KEY_MAX_THREAD_NR, 2, int)

        self.TVB_STORAGE = self.manager.get_attribute(stored.KEY_STORAGE, self.DEFAULT_STORAGE, str)
        # For tests we will place logs in workspace (current folder), to have them visible from Hudson.
        self.TVB_LOG_FOLDER = "TEST_OUTPUT"
        self.TVB_TEMP_FOLDER = os.path.join(self.TVB_STORAGE, "TEMP")

        self.db = DBSettings(self.manager, self.DEFAULT_STORAGE, self.TVB_STORAGE)
        self.db.DB_URL = 'sqlite:///' + os.path.join(self.TVB_STORAGE, "tvb-database.db")
        self.db.SELECTED_DB = 'sqlite'

        self.hpc.CRYPT_DATADIR = self.manager.get_attribute(stored.KEY_CRYPT_DATADIR,
                                                            os.path.join(self.TVB_STORAGE, '.data'))
        self.hpc.CRYPT_PASSDIR = self.manager.get_attribute(stored.KEY_CRYPT_PASSDIR,
                                                            os.path.join(self.TVB_STORAGE, '.pass'))

    def initialize_profile(self, change_logger_in_dev=False):
        super(TestSQLiteProfile, self).initialize_profile(change_logger_in_dev=change_logger_in_dev)

        from tvb.core.utils import get_matlab_executable
        self.MATLAB_EXECUTABLE = get_matlab_executable()

    def initialize_for_deployment(self):
        """
        This profile will not be used in deployment, so we try to avoid problems due to too much initialization
        """
        pass


class TestPostgresProfile(TestSQLiteProfile):
    """
    Defines settings for running tests on a Postgres database.
    """

    def __init__(self):
        super(TestPostgresProfile, self).__init__()

        # Used DB url: IP,PORT. The DB needs to be created in advance.
        self.db.DB_URL = 'postgresql+psycopg2://postgres:root@127.0.0.1:5432/tvb-test?user=postgres&password=postgres'
        self.db.SELECTED_DB = 'postgres'



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import json
from sqlalchemy import Column, Integer, ForeignKey, String, Boolean
from sqlalchemy.orm import relationship
from tvb.adapters.datatypes.db.connectivity import ConnectivityIndex
from tvb.adapters.datatypes.db.region_mapping import RegionVolumeMappingIndex, RegionMappingIndex
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.core.entities.model.model_datatype import DataTypeMatrix
from tvb.core.entities.storage import dao
from tvb.datatypes.graph import Covariance, CorrelationCoefficients, ConnectivityMeasure


class CovarianceIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid), nullable=not Covariance.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    def fill_from_has_traits(self, datatype):
        # type: (Covariance)  -> None
        super(CovarianceIndex, self).fill_from_has_traits(datatype)
        self.fk_source_gid = datatype.source.gid.hex


class CorrelationCoefficientsIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid),
                           nullable=not CorrelationCoefficients.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    labels_ordering = Column(String)

    def get_extra_info(self):
        labels_dict = {}
        labels_dict["labels_ordering"] = self.source.labels_ordering
        labels_dict["labels_dimensions"] = self.source.labels_dimensions
        return labels_dict

    def fill_from_has_traits(self, datatype):
        # type: (CorrelationCoefficients)  -> None
        super(CorrelationCoefficientsIndex, self).fill_from_has_traits(datatype)
        self.labels_ordering = json.dumps(datatype.labels_ordering)
        self.fk_source_gid = datatype.source.gid.hex


class ConnectivityMeasureIndex(DataTypeMatrix):
    id = Column(Integer, ForeignKey(DataTypeMatrix.id), primary_key=True)

    fk_connectivity_gid = Column(String(32), ForeignKey(ConnectivityIndex.gid),
                                 nullable=ConnectivityMeasure.connectivity.required)
    connectivity = relationship(ConnectivityIndex, foreign_keys=fk_connectivity_gid,
                                primaryjoin=ConnectivityIndex.gid == fk_connectivity_gid)

    has_surface_mapping = Column(Boolean, nullable=False, default=False)

    def fill_from_h5(self, h5_file):
        super(ConnectivityMeasureIndex, self).fill_from_h5(h5_file)
        self.fk_connectivity_gid = h5_file.connectivity.load().hex
        self.title = h5_file.title.load()
        self.has_volume_mapping = False
        self.has_surface_mapping = False
        rm_list = dao.get_generic_entity(RegionMappingIndex, self.fk_connectivity_gid, 'fk_connectivity_gid')
        if rm_list:
            self.has_surface_mapping = True

        rvm_list = dao.get_generic_entity(RegionVolumeMappingIndex, self.fk_connectivity_gid, 'fk_connectivity_gid')
        if rvm_list:
            self.has_volume_mapping = True

    def fill_from_has_traits(self, datatype):
        # type: (ConnectivityMeasure)  -> None
        super(ConnectivityMeasureIndex, self).fill_from_has_traits(datatype)
        self.fk_connectivity_gid = datatype.connectivity.gid.hex
        self.title = datatype.title

        self.has_volume_mapping = False
        self.has_surface_mapping = False

        no_reg = datatype.connectivity.number_of_regions
        if not (no_reg in self.parsed_shape):
            return

        rm_list = dao.get_generic_entity(RegionMappingIndex, self.fk_connectivity_gid, 'fk_connectivity_gid')
        if rm_list:
            self.has_surface_mapping = True

        rvm_list = dao.get_generic_entity(RegionVolumeMappingIndex, self.fk_connectivity_gid, 'fk_connectivity_gid')
        if rvm_list:
            self.has_volume_mapping = True


    @property
    def display_name(self):
        """
        Overwrite from superclass and add number of regions field
        """
        result = super(ConnectivityMeasureIndex, self).display_name
        if self.title:
            result = result + " - " + (self.title if len(self.title) < 50 else self.title[:46] + "...")
        return result



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change Project structure for TVB version 1.1.1.
In this version the location where ResultFigures are stored has changed.

.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

from tvb.basic.profile import TvbProfile
from tvb.core.entities.storage import dao
from tvb.core.entities.file.files_helper import FilesHelper
from tvb.core.project_versions.project_update_manager import ProjectUpdateManager


PAGE_SIZE = 20


def _figures_in_project(project_id):
    grouped_figures, _ = dao.get_previews(project_id)
    figures = []

    for fv in grouped_figures.values():
        figures.extend(fv)

    return figures


def update():
    """
    Move images previously stored in TVB operation folders, in a single folder/project.
    """
    projects_count = dao.get_all_projects(is_count=True)

    for page_start in range(0, projects_count, PAGE_SIZE):

        projects_page = dao.get_all_projects(page_start=page_start, page_size=PAGE_SIZE)

        for project in projects_page:
            figures = _figures_in_project(project.id)

            for figure in figures:
                figure.file_path = "%s-%s" % (figure.operation.id, figure.file_path)

            dao.store_entities(figures)

            project_path = FilesHelper().get_project_folder(project)
            update_manager = ProjectUpdateManager(project_path)
            update_manager.run_all_updates()

            project.version = TvbProfile.current.version.PROJECT_VERSION
            dao.store_entity(project)





# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
.. moduleauthor:: Bogdan Valean <bogdan.valean@codemart.ro>
"""

import os
from functools import partial
from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.core.entities.file.data_encryption_handler import encryption_handler
from tvb.core.entities.model.model_operation import STATUS_ERROR, STATUS_CANCELED, STATUS_FINISHED
from tvb.core.entities.model.model_operation import STATUS_STARTED, STATUS_PENDING
from tvb.core.entities.storage import dao
from tvb.core.services.backend_clients.hpc_scheduler_client import HPCSchedulerClient, HPCJobStatus
from tvb.core.services.exceptions import OperationException

try:
    from pyunicore.client import Job, Transport
except ImportError:
    from tvb.basic.config.settings import HPCSettings
    HPCSettings.CAN_RUN_HPC = False


class HPCOperationService(object):
    LOGGER = get_logger(__name__)

    @staticmethod
    def _operation_error(operation):
        operation.mark_complete(STATUS_ERROR)
        dao.store_entity(operation)

    @staticmethod
    def _operation_canceled(operation):
        operation.mark_complete(STATUS_CANCELED)
        dao.store_entity(operation)

    @staticmethod
    def _operation_started(operation):
        operation.start_now()
        dao.store_entity(operation)

    @staticmethod
    def _operation_finished(operation, simulator_gid):
        op_ident = dao.get_operation_process_for_operation(operation.id)
        # TODO: Handle login
        job = Job(Transport(os.environ[HPCSchedulerClient.CSCS_LOGIN_TOKEN_ENV_KEY]),
                  op_ident.job_id)

        operation = dao.get_operation_by_id(operation.id)
        folder = HPCSchedulerClient.file_handler.get_project_folder(operation.project)
        if encryption_handler.encryption_enabled():
            encryption_handler.inc_project_usage_count(folder)
            encryption_handler.sync_folders(folder)

        try:
            sim_h5_filenames, metric_op, metric_h5_filename = \
                HPCSchedulerClient.stage_out_to_operation_folder(job.working_dir, operation, simulator_gid)

            operation.mark_complete(STATUS_FINISHED)
            dao.store_entity(operation)
            HPCSchedulerClient().update_db_with_results(operation, sim_h5_filenames, metric_op, metric_h5_filename)

        except OperationException as exception:
            HPCOperationService.LOGGER.error(exception)
            HPCOperationService._operation_error(operation)

        finally:
            if encryption_handler.encryption_enabled():
                encryption_handler.sync_folders(folder)
                encryption_handler.set_project_inactive(operation.project)

    @staticmethod
    def handle_hpc_status_changed(operation, simulator_gid, new_status):
        # type: (Operation, str, str) -> None

        switcher = {
            STATUS_ERROR: HPCOperationService._operation_error,
            STATUS_CANCELED: HPCOperationService._operation_canceled,
            STATUS_STARTED: HPCOperationService._operation_started,
            STATUS_FINISHED: partial(HPCOperationService._operation_finished, simulator_gid=simulator_gid)
        }
        update_func = switcher.get(new_status, lambda: "Invalid operation status")
        update_func(operation)

    @staticmethod
    def check_operations_job():
        operations = dao.get_operations()
        if operations is None or len(operations) == 0:
            return

        for operation in operations:
            HPCOperationService.LOGGER.info("Start processing operation {}".format(operation.id))
            try:
                op_ident = dao.get_operation_process_for_operation(operation.id)
                if op_ident is not None:
                    transport = Transport(os.environ[HPCSchedulerClient.CSCS_LOGIN_TOKEN_ENV_KEY])
                    job = Job(transport, op_ident.job_id)
                    job_status = job.properties['status']
                    if job.is_running():
                        if operation.status == STATUS_PENDING and job_status == HPCJobStatus.READY.value:
                            HPCOperationService._operation_started(operation)
                        HPCOperationService.LOGGER.info(
                            "CSCS job status: {} for operation {}.".format(job_status, operation.id))
                        return
                    HPCOperationService.LOGGER.info(
                        "Job for operation {} has status {}".format(operation.id, job_status))
                    if job_status == HPCJobStatus.SUCCESSFUL.value:
                        simulator_gid = operation.view_model_gid
                        HPCOperationService._operation_finished(operation, simulator_gid)
                    else:
                        HPCOperationService._operation_error(operation)
            except Exception:
                HPCOperationService.LOGGER.error(
                    "There was an error on background processing process for operation {}".format(operation.id),
                    exc_info=True)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from tvb.simulator.coupling import *
from tvb.adapters.simulator.form_with_ranges import FormWithRanges
from tvb.core.neotraits.forms import ArrayField, BoolField


def get_coupling_to_form_dict():
    coupling_class_to_form = {
        Linear: LinearCouplingForm,
        Scaling: ScalingCouplingForm,
        HyperbolicTangent: HyperbolicTangentCouplingForm,
        Sigmoidal: SigmoidalCouplingForm,
        SigmoidalJansenRit: SigmoidalJansenRitForm,
        PreSigmoidal: PreSigmoidalCouplingForm,
        Difference: DifferenceCouplingForm,
        Kuramoto: KuramotoCouplingForm
    }
    return coupling_class_to_form


def get_ui_name_to_coupling_dict():
    ui_name_to_coupling = {}
    for coupling_class in get_coupling_to_form_dict():
        ui_name_to_coupling.update({coupling_class.__name__: coupling_class})

    return ui_name_to_coupling


def get_form_for_coupling(coupling_class):
    return get_coupling_to_form_dict().get(coupling_class)


class LinearCouplingForm(FormWithRanges):

    def __init__(self):
        super(LinearCouplingForm, self).__init__()
        self.a = ArrayField(Linear.a)
        self.b = ArrayField(Linear.b)


class ScalingCouplingForm(FormWithRanges):

    def __init__(self):
        super(ScalingCouplingForm, self).__init__()
        self.a = ArrayField(Scaling.a)


class HyperbolicTangentCouplingForm(FormWithRanges):

    def __init__(self):
        super(HyperbolicTangentCouplingForm, self).__init__()
        self.a = ArrayField(HyperbolicTangent.a)
        self.b = ArrayField(HyperbolicTangent.b)
        self.midpoint = ArrayField(HyperbolicTangent.midpoint)
        self.sigma = ArrayField(HyperbolicTangent.sigma)


class SigmoidalCouplingForm(FormWithRanges):

    def __init__(self):
        super(SigmoidalCouplingForm, self).__init__()
        self.cmin = ArrayField(Sigmoidal.cmin)
        self.cmax = ArrayField(Sigmoidal.cmax)
        self.midpoint = ArrayField(Sigmoidal.midpoint)
        self.a = ArrayField(Sigmoidal.a)
        self.sigma = ArrayField(Sigmoidal.sigma)


class SigmoidalJansenRitForm(FormWithRanges):

    def __init__(self):
        super(SigmoidalJansenRitForm, self).__init__()
        self.cmin = ArrayField(SigmoidalJansenRit.cmin)
        self.cmax = ArrayField(SigmoidalJansenRit.cmax)
        self.midpoint = ArrayField(SigmoidalJansenRit.midpoint)
        self.r = ArrayField(SigmoidalJansenRit.r)
        self.a = ArrayField(SigmoidalJansenRit.a)


class PreSigmoidalCouplingForm(FormWithRanges):

    def __init__(self):
        super(PreSigmoidalCouplingForm, self).__init__()
        self.H = ArrayField(PreSigmoidal.H)
        self.Q = ArrayField(PreSigmoidal.Q)
        self.G = ArrayField(PreSigmoidal.G)
        self.P = ArrayField(PreSigmoidal.P)
        self.theta = ArrayField(PreSigmoidal.theta)
        self.dynamic = BoolField(PreSigmoidal.dynamic)
        self.globalT= BoolField(PreSigmoidal.globalT)


class DifferenceCouplingForm(FormWithRanges):

    def __init__(self):
        super(DifferenceCouplingForm, self).__init__()
        self.a = ArrayField(Difference.a)


class KuramotoCouplingForm(FormWithRanges):

    def __init__(self):
        super(KuramotoCouplingForm, self).__init__()
        self.a = ArrayField(Kuramoto.a)


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Handlers for reading/writing from/into OBJ format Surfaces are defined in this module.

.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

from tvb.basic.logger.builder import get_logger



class ObjParser(object):
    """
    This class reads geometry from a simple wavefront obj file.
    ``self.vertices``, ``self.tex_coords``, ``self.normals`` are lists of vectors represented as tuples
    ``self.faces`` is a list of faces. A face is a list of vertex info.
    Vertex info is a tuple vertex_index, tex_index, normal_index.
    """


    def __init__(self):
        self.logger = get_logger(__name__)
        self.vertices = []
        self.normals = []
        self.tex_coords = []
        self.faces = []


    def parse_v(self, args):
        self.vertices.append(tuple(float(x) for x in args[:3]))


    def parse_vn(self, args):
        self.normals.append(tuple(float(x) for x in args[:3]))


    def parse_vt(self, args):
        self.tex_coords.append(tuple(float(x) for x in args[:3]))


    def parse_f(self, args):
        if len(args) < 3:
            raise ValueError("Require at least 3 vertices per face")

        face = []

        for v in args:
            indices = []

            for j in v.split('/'):
                if j == '':
                    indices.append(-1)
                else:
                    indices.append(int(j) - 1)

            while len(indices) < 3:
                indices.append(-1)

            face.append(indices)
        self.faces.append(face)


    def read(self, obj_file):
        line_nr = -1
        try:
            for line_nr, line in enumerate(obj_file):
                line = line.strip()
                if line == "" or line[0] == '#':
                    continue
                tokens = line.split()
                data_type = tokens[0]
                datatype_parser = getattr(self, "parse_%s" % data_type, None)

                if datatype_parser is not None:
                    datatype_parser(tokens[1:])
                else:
                    self.logger.warning("Unsupported token type %s" % data_type)
        except ValueError as ex:
            raise ValueError("%s at line %d" % (ex, line_nr))



class ObjWriter(object):

    def __init__(self, obj_file):
        self.logger = get_logger(__name__)
        self.file = obj_file


    def _write_vector(self, type_code, v):
        v_str = ' '.join(str(e) for e in v)
        self.file.write('%s %s\n' % (type_code, v_str))


    def _write_face(self, v_info, write_normals):
        fs = []
        for info in v_info:
            info += 1
            if not write_normals:
                s = '%d' % info
            else:
                s = '%d//%d' % (info, info)
            fs.append(s)
        self.file.write('f %s \n' % ' '.join(fs))


    def write(self, vertices, faces, normals=None, comment=''):
        """
        :param vertices:, :param normals: are lists of vectors or ndarrays of shape (n,3)
        :param faces: A face is a list of 3 vertex indices.
        Normal indices not supported. Texture uv's not supported.
        This method does not yet validate the input, so send valid data.
        """
        self.file.write('# %s\n' % comment)
        for v in vertices:
            self._write_vector('v', v)

        self.file.write('\n')

        if normals is not None:
            for v in normals:
                self._write_vector('vn', v)
            self.file.write('\n')

        for f in faces:
            self._write_face(f, normals is not None)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
"""

import copy
import json
import shutil
import uuid
import numpy

from tvb.basic.logger.builder import get_logger
from tvb.core.entities import load
from tvb.core.entities.file.files_helper import FilesHelper
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.entities.model.model_datatype import DataTypeGroup
from tvb.core.entities.storage import dao
from tvb.core.neocom import h5
from tvb.core.services.algorithm_service import AlgorithmService
from tvb.core.services.burst_service import BurstService
from tvb.core.services.exceptions import BurstServiceException
from tvb.core.services.operation_service import OperationService
from tvb.simulator.integrators import IntegratorStochastic


class SimulatorService(object):
    def __init__(self):
        self.logger = get_logger(self.__class__.__module__)
        self.burst_service = BurstService()
        self.operation_service = OperationService()
        self.algorithm_service = AlgorithmService()
        self.files_helper = FilesHelper()

    @staticmethod
    def _reset_model(session_stored_simulator):
        session_stored_simulator.model = type(session_stored_simulator.model)()
        vi_indexes = session_stored_simulator.determine_indexes_for_chosen_vars_of_interest()
        vi_indexes = numpy.array(list(vi_indexes.values()))
        for monitor in session_stored_simulator.monitors:
            monitor.variables_of_interest = vi_indexes

    def reset_at_connectivity_change(self, is_simulator_copy, form, session_stored_simulator):
        """
        In case the user copies a simulation and changes the Connectivity, we want to reset the Model and Noise
        parameters because they might not fit to the new Connectivity's nr of regions.
        """
        if is_simulator_copy and form.connectivity.value != session_stored_simulator.connectivity:
            self._reset_model(session_stored_simulator)
            if issubclass(type(session_stored_simulator.integrator), IntegratorStochastic):
                session_stored_simulator.integrator.noise = type(session_stored_simulator.integrator.noise)()

    def reset_at_surface_change(self, is_simulator_copy, form, session_stored_simulator):
        """
        In case the user copies a surface-simulation and changes the Surface, we want to reset the Model
        parameters because they might not fit to the new Surface's nr of vertices.
        """
        if is_simulator_copy and (session_stored_simulator.surface is None and form.surface.value
                                  or session_stored_simulator.surface and
                                  form.surface.value != session_stored_simulator.surface.surface_gid):
            self._reset_model(session_stored_simulator)

    @staticmethod
    def _set_simulator_range_parameter(simulator, range_parameter_name, range_parameter_value):
        range_param_name_list = range_parameter_name.split('.')
        current_attr = simulator
        for param_name in range_param_name_list[:len(range_param_name_list) - 1]:
            current_attr = getattr(current_attr, param_name)
        setattr(current_attr, range_param_name_list[-1], range_parameter_value)

    def async_launch_and_prepare_simulation(self, burst_config, user, project, simulator_algo, simulator):
        try:
            operation = self.operation_service.prepare_operation(user.id, project, simulator_algo,
                                                                 view_model=simulator, burst_gid=burst_config.gid,
                                                                 op_group_id=burst_config.fk_operation_group)
            burst_config = self.burst_service.update_simulation_fields(burst_config, operation.id, simulator.gid)
            storage_path = self.files_helper.get_project_folder(project, str(operation.id))
            self.burst_service.store_burst_configuration(burst_config, storage_path)

            wf_errs = 0
            try:
                OperationService().launch_operation(operation.id, True)
                return operation
            except Exception as excep:
                self.logger.error(excep)
                wf_errs += 1
                if burst_config:
                    self.burst_service.mark_burst_finished(burst_config, error_message=str(excep))

            self.logger.debug("Finished launching workflow. The operation was launched successfully, " +
                              str(wf_errs) + " had error on pre-launch steps")

        except Exception as excep:
            self.logger.error(excep)
            if burst_config:
                self.burst_service.mark_burst_finished(burst_config, error_message=str(excep))

    def prepare_simulation_on_server(self, user_id, project, algorithm, zip_folder_path, simulator_file):
        simulator_vm = h5.load_view_model_from_file(simulator_file)
        operation = self.operation_service.prepare_operation(user_id, project, algorithm, view_model=simulator_vm)
        self.async_launch_simulation_on_server(operation, zip_folder_path)

        return operation

    def async_launch_simulation_on_server(self, operation, zip_folder_path):
        try:
            OperationService().launch_operation(operation.id, True)
            return operation
        except Exception as excep:
            self.logger.error(excep)
        finally:
            shutil.rmtree(zip_folder_path)

    @staticmethod
    def _set_range_param_in_dict(param_value):
        if type(param_value) is numpy.ndarray:
            return param_value[0]
        elif isinstance(param_value, uuid.UUID):
            return param_value.hex
        else:
            return param_value

    def async_launch_and_prepare_pse(self, burst_config, user, project, simulator_algo, range_param1, range_param2,
                                     session_stored_simulator):
        try:
            algo_category = simulator_algo.algorithm_category
            operation_group = burst_config.operation_group
            metric_operation_group = burst_config.metric_operation_group
            operations = []
            range_param2_values = [None]
            if range_param2:
                range_param2_values = range_param2.get_range_values()
            first_simulator = None

            for param1_value in range_param1.get_range_values():
                for param2_value in range_param2_values:
                    # Copy, but generate a new GUID for every Simulator in PSE
                    simulator = copy.deepcopy(session_stored_simulator)
                    simulator.gid = uuid.uuid4()
                    self._set_simulator_range_parameter(simulator, range_param1.name, param1_value)

                    ranges = {range_param1.name: self._set_range_param_in_dict(param1_value)}

                    if param2_value is not None:
                        self._set_simulator_range_parameter(simulator, range_param2.name, param2_value)
                        ranges[range_param2.name] = self._set_range_param_in_dict(param2_value)

                    ranges = json.dumps(ranges)

                    operation = self.operation_service.prepare_operation(user.id, project, simulator_algo,
                                                                         view_model=simulator, ranges=ranges,
                                                                         burst_gid=burst_config.gid,
                                                                         op_group_id=burst_config.fk_operation_group)
                    simulator.range_values = ranges
                    operations.append(operation)
                    if first_simulator is None:
                        first_simulator = simulator

            first_operation = operations[0]
            storage_path = self.files_helper.get_project_folder(project, str(first_operation.id))
            burst_config = self.burst_service.update_simulation_fields(burst_config, first_operation.id,
                                                                       first_simulator.gid)
            self.burst_service.store_burst_configuration(burst_config, storage_path)
            datatype_group = DataTypeGroup(operation_group, operation_id=first_operation.id,
                                           fk_parent_burst=burst_config.gid,
                                           state=algo_category.defaultdatastate)
            dao.store_entity(datatype_group)

            metrics_datatype_group = DataTypeGroup(metric_operation_group, fk_parent_burst=burst_config.gid,
                                                   state=algo_category.defaultdatastate)
            dao.store_entity(metrics_datatype_group)

            wf_errs = 0
            for operation in operations:
                try:
                    OperationService().launch_operation(operation.id, True)
                except Exception as excep:
                    self.logger.error(excep)
                    wf_errs += 1
                    self.burst_service.mark_burst_finished(burst_config, error_message=str(excep))

            self.logger.debug("Finished launching workflows. " + str(len(operations) - wf_errs) +
                              " were launched successfully, " + str(wf_errs) + " had error on pre-launch steps")
            return first_operation

        except Exception as excep:
            self.logger.error(excep)
            self.burst_service.mark_burst_finished(burst_config, error_message=str(excep))

    @staticmethod
    def compute_conn_branch_conditions(is_branch, simulator):
        if not is_branch:
            return None

        conn = load.load_entity_by_gid(simulator.connectivity)
        if conn.number_of_regions:
            return FilterChain(fields=[FilterChain.datatype + '.number_of_regions'],
                               operations=["=="], values=[conn.number_of_regions])

    @staticmethod
    def validate_first_fragment(form, project_id, conn_idx):
        conn_count = dao.count_datatypes(project_id, conn_idx)
        if conn_count == 0:
            form.connectivity.errors.append("No connectivity in the project! Simulation cannot be started without "
                                            "a connectivity!")

    def get_simulation_state_index(self, burst_config, simulation_history_class):
        parent_burst = burst_config.parent_burst_object
        simulation_state_index = dao.get_generic_entity(simulation_history_class, parent_burst.gid, "fk_parent_burst")

        if simulation_state_index is None or len(simulation_state_index) < 1:
            exc = BurstServiceException("Simulation State not found for %s, thus we are unable to branch from "
                                        "it!" % burst_config.name)
            self.logger.error(exc)
            raise exc

        return simulation_state_index
